/* SPDX-License-Identifier: GPL-2.0 or Linux-OpenIB */
/* Copyright (c) 2019 Intel Corporation */
#ifndef IRDMA_KCOMPAT_H
#define IRDMA_KCOMPAT_H

#ifndef LINUX_VERSION_CODE
#include <linux/version.h>
#else
#define KERNEL_VERSION(a, b, c) (((a) << 16) + ((b) << 8) + (c))
#endif
#include <linux/init.h>
#include <linux/types.h>
#include <linux/errno.h>
#include <linux/module.h>
#include <linux/pci.h>
#include <linux/netdevice.h>
#include <linux/etherdevice.h>
#include <linux/skbuff.h>
#include <linux/ioport.h>
#include <linux/slab.h>
#include <linux/list.h>
#include <linux/delay.h>
#include <linux/sched.h>
#include <linux/in.h>
#include <linux/ip.h>
#include <linux/udp.h>
#include <linux/mii.h>
#include <linux/vmalloc.h>
#include <linux/irq.h>
#include <linux/hugetlb.h>
#include <asm/io.h>
#include <linux/ethtool.h>
#include <linux/if_vlan.h>
#include <net/ipv6.h>
#include <net/ip6_route.h>
#include <net/route.h>
#include <rdma/ib_cache.h>
#include <rdma/ib_umem.h>
#if IS_ENABLED(CONFIG_NET_DEVLINK)
#include <net/devlink.h>
#endif
#if LINUX_VERSION_CODE >= KERNEL_VERSION(4, 14, 0)
#include <rdma/uverbs_ioctl.h>
#endif
#if LINUX_VERSION_CODE >= KERNEL_VERSION(3, 4, 0)
#include <linux/kconfig.h>
#endif
#if LINUX_VERSION_CODE >= KERNEL_VERSION(4, 19, 0)
#include <net/secure_seq.h>
#endif
#if LINUX_VERSION_CODE < KERNEL_VERSION(4, 4, 0)
#include <asm-generic/io-64-nonatomic-lo-hi.h>
#else
#include <linux/io-64-nonatomic-lo-hi.h>
#endif

#include "distro_ver.h"

#ifdef __OFED_BUILD__
#include "ofed_kcompat.h"
#elif defined(RHEL_RELEASE_CODE)
#include "rhel_kcompat.h"
#elif defined(CONFIG_SUSE_KERNEL)
#include "suse_kcompat.h"
#else
#include "linux_kcompat.h"
#endif

/* Mapping IRDMA driver ID to I40IW till we are in k.org */
#define RDMA_DRIVER_IRDMA RDMA_DRIVER_I40IW

#if (LINUX_VERSION_CODE < KERNEL_VERSION(4, 14, 0))
#define TIMER_DATA_TYPE		unsigned long
#define TIMER_FUNC_TYPE		void (*)(TIMER_DATA_TYPE)

#define timer_setup(timer, callback, flags)				\
	__setup_timer((timer), (TIMER_FUNC_TYPE)(callback),		\
		      (TIMER_DATA_TYPE)(timer), (flags))

#define from_timer(var, callback_timer, timer_fieldname) \
	container_of(callback_timer, typeof(*var), timer_fieldname)
#endif /* (LINUX_VERSION_CODE < KERNEL_VERSION(4, 14, 0)) */

#if LINUX_VERSION_CODE >= KERNEL_VERSION(5, 5, 0)
/* https://lore.kernel.org/linux-rdma/20191217210406.GC17227@ziepe.ca/
 * This series adds mmap DB support and also extends rdma_user_mmap_io API
 * with an extra param
 */
#define RDMA_MMAP_DB_SUPPORT
#define DEVLINK_RELOAD_DOWN_NETNS_CHANGE
#endif

#ifndef __OFED_BUILD__
#if LINUX_VERSION_CODE < KERNEL_VERSION(5, 0, 0)
#define dma_alloc_coherent dma_zalloc_coherent
#endif
#endif

#if LINUX_VERSION_CODE < KERNEL_VERSION(5, 0, 0)
#define IB_GET_NETDEV_OP_NOT_DEPRECATED
#endif

struct dst_entry *irdma_get_fl6_dst(struct sockaddr_in6 *, struct sockaddr_in6 *);
struct neighbour *irdma_get_neigh_ipv6(struct dst_entry *, struct sockaddr_in6 *);
struct neighbour *irdma_get_neigh_ipv4(struct rtable *, __be32 *);
void kc_set_loc_seq_num_mss(struct irdma_cm_node *cm_node);

/*************************DEVLINK*********************************************/
#if LINUX_VERSION_CODE >= KERNEL_VERSION(5, 4, 0)
#define IRDMA_DEVLINK_RELOAD_SPLIT_OPS
#endif
#if IS_ENABLED(CONFIG_NET_DEVLINK)
#ifdef __DEVLINK_PARAM_MAX_STRING_VALUE
#define DEVLINK_SUPPORTED
#endif
#endif
/*********************************************************/

/*********************************************************/
#ifdef IRDMA_SET_DRIVER_ID
#define kc_set_driver_id(ibdev) ibdev.driver_id = RDMA_DRIVER_I40IW
#else
#define kc_set_driver_id(x)
#endif /* IRDMA_SET_DRIVER_ID */
/*****************************************************************************/

#ifndef ether_addr_copy
#define ether_addr_copy(mac_addr, new_mac_addr) memcpy(mac_addr, new_mac_addr, ETH_ALEN)
#endif
#ifndef eth_zero_addr
#define eth_zero_addr(mac_addr) memset(mac_addr, 0x00, ETH_ALEN)
#endif

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 35)
#define irdma_for_each_ipv6_addr(ifp, tmp, idev) list_for_each_entry_safe(ifp, tmp, &idev->addr_list, if_list)
#else
#define irdma_for_each_ipv6_addr(ifp, tmp, idev) for (ifp = idev->addr_list; ifp != NULL; ifp = ifp->if_next)
#endif /* >= 2.6.35 */

#if LINUX_VERSION_CODE < KERNEL_VERSION(3, 9, 0)
#define netdev_master_upper_dev_get irdma_netdev_master_upper_dev_get
struct net_device *irdma_netdev_master_upper_dev_get(struct net_device *);

#if LINUX_VERSION_CODE < KERNEL_VERSION(3, 1, 0)
#define neigh_release(neigh)
#endif /* < 3.1.0 */

#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 39)
#define ip_route_output irdma_ip_route_output
struct rtable *irdma_ip_route_output(struct net *, __be32, __be32, u8, int);
#endif /* < 2.6.39 */
#endif /* < 3.9.0 */

#ifdef IB_FW_VERSION_NAME_MAX
void irdma_get_dev_fw_str(struct ib_device *dev, char *str);
#else
void irdma_get_dev_fw_str(struct ib_device *dev, char *str, size_t str_len);
#endif /* IB_FW_VERSION_NAME_MAX */

/*****************************************************************************/

struct dst_entry *irdma_get_fl6_dst(struct sockaddr_in6 *src_addr,
				    struct sockaddr_in6 *dst_addr);
struct neighbour *irdma_get_neigh_ipv6(struct dst_entry *dst,
				       struct sockaddr_in6 *dst_ipaddr);
struct neighbour *irdma_get_neigh_ipv4(struct rtable *rt, __be32 *dst_ipaddr);

#ifdef CREATE_AH_VER_5
int irdma_create_ah_v2(struct ib_ah *ib_ah,
		       struct rdma_ah_attr *attr, u32 flags,
		       struct ib_udata *udata);
int irdma_create_ah(struct ib_ah *ibah,
		    struct rdma_ah_init_attr *attr,
		    struct ib_udata *udata);
#endif

#ifdef CREATE_AH_VER_4
struct ib_ah *irdma_create_ah(struct ib_pd *ibpd,
			      struct rdma_ah_attr *attr,
			      struct ib_udata *udata);
#endif

#ifdef CREATE_AH_VER_3
struct ib_ah *irdma_create_ah(struct ib_pd *ibpd,
			      struct rdma_ah_attr *attr,
			      u32 flags,
			      struct ib_udata *udata);
#endif

#ifdef CREATE_AH_VER_2
int irdma_create_ah(struct ib_ah *ib_ah,
		    struct rdma_ah_attr *attr, u32 flags,
		    struct ib_udata *udata);
#endif

#ifdef CREATE_AH_VER_1
struct ib_ah *irdma_create_ah(struct ib_pd *ibpd,
#if defined(RHEL_RELEASE_CODE)
#if (defined(RHEL_7_4) && !defined(__OFED_BUILD__))
			      struct ib_ah_attr *attr,
#else
			      struct rdma_ah_attr *attr,
#endif
#elif !defined(__OFED_BUILD__) && !defined(CONFIG_SUSE_KERNEL) && (LINUX_VERSION_CODE < KERNEL_VERSION(4, 12, 0))
			      struct ib_ah_attr *attr,
#else
			      struct rdma_ah_attr *attr,
#endif
			      struct ib_udata *udata);
#endif

#ifdef DESTROY_AH_VER_3
void irdma_destroy_ah(struct ib_ah *ibah, u32 flags);
#endif

#ifdef DESTROY_AH_VER_2
int irdma_destroy_ah(struct ib_ah *ibah, u32 flags);
#endif

#ifdef DESTROY_AH_VER_1
int irdma_destroy_ah(struct ib_ah *ibah);
#endif

void irdma_unregister_rdma_device(struct ib_device *ibdev);
#ifndef RDMA_MMAP_DB_SUPPORT
int rdma_user_mmap_io(struct ib_ucontext *ucontext, struct vm_area_struct *vma,
		      unsigned long pfn, unsigned long size, pgprot_t prot);
#endif
void irdma_disassociate_ucontext(struct ib_ucontext *context);
int kc_irdma_set_roce_cm_info(struct irdma_qp *iwqp,
			      struct ib_qp_attr *attr,
			      u16 *vlan_id);
int kc_irdma_create_sysfs_file(struct ib_device *ibdev);
struct irdma_device *kc_irdma_get_device(struct net_device *netdev);
void kc_irdma_put_device(struct irdma_device *iwdev);
#endif /* IRDMA_KCOMPAT_H_ */
